from __future__ import annotations

import os
import yaml
from typing import Dict, List
from pathlib import Path

from yaml_extender import yaml_loader
from yaml_extender.resolver.include_resolver import IncludeResolver
from yaml_extender.resolver.inline_loop_resolver import InlineLoopResolver
from yaml_extender.resolver.loop_resolver import LoopResolver
from yaml_extender.resolver.reference_resolver import ReferenceResolver

ENV_KEY = "env"
PARAM_KEY = "param"


class XYmlFile:

    def __init__(self, filepath: Path, params: Dict = None, include_dirs: List[Path] | None = None):
        self.params = params
        if include_dirs:
            self.include_dirs: List[Path] = include_dirs
        else:
            self.include_dirs: List[Path] = []
        self.filepath = filepath.absolute()
        self.root_dir = filepath.parent
        # Use root_dir and cwd as default include paths.
        if self.root_dir not in self.include_dirs:
            self.include_dirs.append(self.root_dir)
        if Path.cwd() not in self.include_dirs:
            self.include_dirs.append(Path.cwd())
        self.content = yaml_loader.load(str(self.filepath))
        self.content = self.resolve()

    def __repr__(self):
        return yaml.dump(self.content)

    def resolve(self):
        inc_resolver = IncludeResolver(self.include_dirs, False)
        processed_content = inc_resolver.resolve(self.content)
        loop_resolver = LoopResolver(False)
        processed_content = loop_resolver.resolve(processed_content)
        inline_loop_resolver = InlineLoopResolver(False)
        processed_content = inline_loop_resolver.resolve(processed_content)
        # Extend config for resolution by ENV and PARAM statements
        config = processed_content.copy()
        config["xyml"] = {}
        config["xyml"][ENV_KEY] = os.environ
        config["xyml"][PARAM_KEY] = self.params
        ref_resolver = ReferenceResolver(False)
        processed_content = ref_resolver.resolve(processed_content, config)
        return processed_content

    def save(self, path: str, sort_keys=False):
        with open(path, 'w') as file:
            yaml.dump(self.content, file, sort_keys=sort_keys)





from __future__ import annotations

import os
import re
from pathlib import Path
from typing import Any, List

from yaml_extender.resolver.reference_resolver import ReferenceResolver
from yaml_extender.resolver.resolver import Resolver
from yaml_extender.xyml_exception import ExtYamlError, ExtYamlSyntaxError
import yaml_extender.logger as logger
import yaml_extender.yaml_loader as yaml_loader


INCLUDE_REGEX = r'([^<]+)\s*(?:<<(.*)>>)?'
INCLUDE_KEY = "xyml.include"


class IncludeResolver(Resolver):

    def __init__(self, include_dirs: List[Path] | None = None, fail_on_resolve: bool = True):
        if include_dirs:
            self.include_dirs: List[Path] = [inc.absolute() for inc in include_dirs]
        else:
            self.include_dirs: List[Path] = []
        if Path.cwd() not in self.include_dirs:
            self.include_dirs.append(Path.cwd())
        super().__init__(fail_on_resolve)

    def _Resolver__resolve(self, cur_value: Any, config: dict) -> dict:
        return self.__resolve_inc(cur_value, config)

    def __resolve_inc(self, cur_value: Any, config: dict) -> Any:
        """
        Resolves all include statements in value

            Returns:
                The content of the original file with all includes resolved.
        """
        if isinstance(cur_value, dict):
            for k, v in list(cur_value.items()):
                if k != INCLUDE_KEY:
                    cur_value[k] = self.__resolve_inc(cur_value[k], config)
                else:
                    include_content = self.__resolve_include_statement(cur_value[INCLUDE_KEY], config)
                    if isinstance(include_content, dict):
                        self.update_content_with_include_content(cur_value, include_content)
                        del cur_value[INCLUDE_KEY]
                    else:
                        return include_content
        elif isinstance(cur_value, list):
            new_content = []
            for i, x in enumerate(cur_value):
                new_value = (self.__resolve_inc(x, config))
                if isinstance(new_value, list):
                    new_content.extend(new_value)
                else:
                    new_content.append(new_value)
            if new_content:
                cur_value = new_content
        return cur_value

    def __resolve_include_statement(self, value: List | str, config: dict) -> dict:
        """Resolves an include statement and return the content"""
        if not isinstance(value, list):
            statements = [value]
        else:
            statements = value
        # Resolve all references in statement
        ref_resolver = ReferenceResolver(False)
        inc_contents = None
        for statement in statements:
            # Resolve include parameters
            match = re.match(INCLUDE_REGEX, statement)
            # Resolve references in filenames
            inc_file_path = ref_resolver.resolve(match.group(1), config)
            logger.info(f"Resolving Include '{inc_file_path}'")
            inc_content = self.__read_included_yaml(inc_file_path)
            # Resolve parameters in included file
            if match.group(2):
                parameters = self.__parse_include_parameters(match.group(2))
                inc_content = ref_resolver.resolve(inc_content, parameters)
            # Add include content to current content
            include_dirs = self.include_dirs.copy()
            include_dirs.append(Path(inc_file_path).parent)
            inc_resolver = IncludeResolver(include_dirs, self.fail_on_resolve)
            inc_content = inc_resolver.__resolve_inc(inc_content, config)
            inc_contents = self.update_inc_content(inc_contents, inc_content)
        return inc_contents

    def update_content_with_include_content(self, existing_content, include_content):
        for k, v in include_content.items():
            if k in existing_content:
                if isinstance(v, dict):
                    self.update_content_with_include_content(existing_content[k], v)
            else:
                existing_content[k] = v

    def update_inc_content(self, content, include):
        """Adds include content to existing content based on current datatype"""
        if content is None:
            return include
        if isinstance(include, list):
            if isinstance(content, dict):
                content = [content]
            content.extend(include)
        elif isinstance(include, dict):
            if isinstance(content, list):
                content.append(include)
            else:
                content.update(include)
        else:
            raise ExtYamlSyntaxError("Resolved include content is not of list or dict type.")
        return content

    def __parse_include_parameters(self, param_string: str) -> dict:
        """Parses an include parameter string into a dict"""
        parameters = {}
        for param in param_string.split(","):
            key, value = [x.strip() for x in param.split("=", maxsplit=1)]
            if not key or not value:
                raise ExtYamlSyntaxError(f"Invalid parameter string {param_string}")
            parameters[key] = yaml_loader.parse_any_value(value)
        return parameters

    def __read_included_yaml(self, file_path: str):
        # Try path with all include dirs respecting the order
        file = Path(file_path)
        if not file.is_absolute():
            for path in self.include_dirs:
                file = path / file_path
                if file.is_file():
                    return yaml_loader.load(str(file))
        else:
            return yaml_loader.load(str(file))
        raise ExtYamlError(f"Include file '{file_path}' not found. Are include directories provided?")



from __future__ import annotations

import re
from pathlib import Path
from typing import Any, Optional

from yaml_extender import yaml_loader
from yaml_extender.resolver.resolver import Resolver
from yaml_extender.xyml_exception import RecursiveReferenceError, ReferenceNotFoundError, ExtYamlSyntaxError

REFERENCE_REGEX = r'\{\{(.+?)(?::(.*?))?\}\}'
ARRAY_REGEX = r'(.*)?\[(\d*)\]'
LIST_FLATTEN_CHARACTER = " "

MAXIMUM_REFERENCE_DEPTH = 30


class ArithmeticOperation:

    SUPPORTED_FUNCS = {"+": lambda x, y: x + y,
                       "-": lambda x, y: x - y,
                       "*": lambda x, y: x * y,
                       "/": lambda x, y: x / y}
    ARITHMETIC_REGEX = r'(.+)([' + ''.join(["\\" + k for k in SUPPORTED_FUNCS.keys()]) + r'])\s*(\d+)'

    def __init__(self, reference: str, operation: str, value: str):
        self.reference = reference.strip()
        self.operation = operation.strip()
        self.value = yaml_loader.parse_numeric_value(value.strip())

    def __repr__(self):
        return f"{self.reference} {self.operation} {self.value}"

    def apply(self, value):
        num_val = yaml_loader.parse_numeric_value(value)
        return ArithmeticOperation.SUPPORTED_FUNCS[self.operation](self.value, num_val)

    @staticmethod
    def parse(expression: str) -> Optional[ArithmeticOperation]:
        match = re.search(ArithmeticOperation.ARITHMETIC_REGEX, expression)
        if match:
            return ArithmeticOperation(match[1], match[2], match[3])
        else:
            return None


class ReferenceResolver(Resolver):

    def __init__(self, fail_on_resolve: bool = True):
        super().__init__(fail_on_resolve)

    def _Resolver__resolve(self, cur_value: Any, config: dict):
        """Resolves all references in a given value using the provided content dict"""
        new_value = cur_value
        if isinstance(cur_value, dict):
            for k in cur_value.keys():
                cur_value[k] = self._Resolver__resolve(cur_value[k], config)
        elif isinstance(cur_value, list):
            new_list = []
            for i, x in enumerate(cur_value):
                resolved_value = self._Resolver__resolve(x, config)
                if isinstance(resolved_value, list):
                    # If the returned value is also a list, extend the current list with it
                    new_list.extend(resolved_value)
                else:
                    new_list.append(self._Resolver__resolve(x, config))
            new_value = new_list
        else:
            new_value = self.resolve_reference(cur_value, config)
        return new_value

    def resolve_reference(self, value: Any, config: dict, depth: int = 0) -> Any:
        if not isinstance(value, str) or "{" not in value:
            return value
        if depth > 30:
            raise RecursiveReferenceError(value)
        new_value = value
        # In order to store the full match the whole regex is packed into a group
        for ref_match in re.finditer(REFERENCE_REGEX, value):
            ref = ref_match.group(1).strip()
            default_value = ref_match.group(2)
            if default_value:
                default_value = yaml_loader.parse_any_value(default_value.strip())
            # Resolve arithmetic operation
            operation = ArithmeticOperation.parse(ref)
            if operation:
                ref = operation.reference
            # Resolve reference, including subrefs
            try:
                ref_val = self.resolve_subrefs(ref, config)
            except ReferenceNotFoundError as ref_err:
                if default_value is not None:
                    ref_val = default_value
                elif self.fail_on_resolve:
                    raise ref_err
                else:
                    ref_val = None

            if ref_val is not None:
                if operation:
                    ref_val = operation.apply(ref_val)
                # If resolved value is of type list, flatten it
                if isinstance(ref_val, list):
                    ref_val = LIST_FLATTEN_CHARACTER.join(ref_val)
                if ref_match.group(0) == value:
                    # If the whole string is just a reference return the value without string replacement
                    # in order to preserve float & int types
                    return ref_val
                else:
                    # Replace the reference string within the value
                    new_value = new_value.replace(ref_match.group(0), str(ref_val))

        if new_value == value:
            if self.fail_on_resolve:
                raise ReferenceNotFoundError(value)
            else:
                return value

        # Resolve recursive references
        new_value = self.resolve_reference(new_value, config, depth + 1)
        return new_value

    def resolve_subrefs(self, fullref: str, current_config: dict):
        if not fullref:
            return current_config
        if "." in fullref:
            ref, sub_ref = fullref.split(".", maxsplit=1)
        else:
            ref = fullref
            sub_ref = None
        # If subref is specifying more than config can resolve, e.g. for include parameter dicts
        # And the resolved value is another reference, append the subref and resolve later
        if isinstance(current_config, str):
            match = re.match(REFERENCE_REGEX, current_config)
            if match:
                # If the current config represents another reference and there are more subrefs specified
                # then extend the reference by the remaining subref
                current_config = match.group(1).strip()
                if match.group(2):
                    current_config += f":{match.group(2)}"
                return "{{" + current_config + f".{fullref}" + "}}"
            else:
                # Fail, because the reference specifies more than can be resolved
                raise ReferenceNotFoundError(fullref)
        elif isinstance(current_config, list):
            if ref.isdigit():
                if len(current_config) > int(ref):
                    current_config = current_config[int(ref)]
                else:
                    raise ReferenceNotFoundError(fullref, ref)
            else:
                # Resolve list of dicts
                value_list = []
                for elem in current_config:
                    try:
                        value_list.append(self.resolve_subrefs(fullref, elem))
                    except ReferenceNotFoundError:
                        pass
                return value_list
        else:
            if ref in current_config:
                current_config = current_config[ref]
            else:
                # Fail, because the reference cannot be found in config
                raise ReferenceNotFoundError(fullref)
        return self.resolve_subrefs(sub_ref, current_config)



from __future__ import annotations

import copy
import re
from pathlib import Path
from typing import Any

from yaml_extender.resolver.reference_resolver import ReferenceResolver
from yaml_extender.resolver.resolver import Resolver
from yaml_extender.xyml_exception import RecursiveReferenceError, ReferenceNotFoundError, ExtYamlSyntaxError

LOOP_KEY = "xyml.for"
INLINE_LOOP_REGEX = r'(\{\{\s*xyml\.for\s*:\s*([^:]+)\s*:\s*([^:]+)\s*:(.+)\}\})'
MAXIMUM_REFERENCE_DEPTH = 30


class InlineLoopResolver(Resolver):

    def __init__(self, fail_on_resolve: bool = True):
        super().__init__(fail_on_resolve)
        self.ref_resolver = ReferenceResolver(False)

    def _Resolver__resolve(self, cur_value: Any, config: dict):
        """Resolves all references in a given value using the provided content dict"""
        new_value = cur_value
        if isinstance(cur_value, dict):
            for k in cur_value.keys():
                cur_value[k] = self._Resolver__resolve(cur_value[k], config)
        elif isinstance(cur_value, list):
            new_list = []
            for i, x in enumerate(cur_value):
                resolved_value = self._Resolver__resolve(x, config)
                if isinstance(resolved_value, list):
                    # If the returned value is also a list, extend the current list with it
                    new_list.extend(resolved_value)
                else:
                    new_list.append(self._Resolver__resolve(x, config))
            new_value = new_list
        elif isinstance(cur_value, str):
            new_value = self.resolve_inline_loop(cur_value, config)
        return new_value

    def resolve_inline_loop(self, value: str, config: dict):
        new_value = value
        for match in re.findall(INLINE_LOOP_REGEX, value):
            full_match = match[0]
            iterator = match[1]
            iteration_value = match[2]
            content = match[3]
            iter_content = config[iteration_value]
            if not isinstance(iter_content, list):
                raise ExtYamlSyntaxError(f"{iteration_value} is not iterable and therefore cannot be used in a loop.")
            new_content = self.get_loop_content(content, iterator, iter_content)
            new_value = new_value.replace(full_match, new_content)
        return new_value

    def get_loop_content(self, content: str, iterator: str, iteration_value: list):
        new_content = ""
        for item in iteration_value:
            new_content += str(self.ref_resolver.resolve(content, {iterator: item}))
        return new_content



class ExtYamlError(Exception):
    pass


class ReferenceNotFoundError(ExtYamlError):

    def __init__(self, reference: str, subref: str = ""):
        self.reference = reference
        self.subref = subref
        self.message = f"Unable to resolve {reference}"
        if self.subref:
            self.message += f" specified sub value {self.subref} not found."


class ExtYamlSyntaxError(ExtYamlError):
    pass


class RecursiveReferenceError(RecursionError):

    def __init__(self, reference):
        self.message = f"Maximum recursive depth reached, while resolving {reference}." \
                       f"Is there a loop in your configuration?"



import logging

LOGGER: logging.Logger = None


def get_logger():
    if not LOGGER:
        init_basic_logger()
    return LOGGER


def debug(msg: str, **kwargs):
    get_logger().debug(msg, **kwargs)


def info(msg: str, **kwargs):
    get_logger().info(msg, **kwargs)


def warning(msg: str, **kwargs):
    get_logger().warning(msg, **kwargs)


def error(msg: str, **kwargs):
    get_logger().error(msg, **kwargs)


def init_basic_logger():
    global LOGGER
    LOGGER = logging.getLogger('xyaml_parser')
    LOGGER.setLevel(logging.DEBUG)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s: [%(levelname)s]: %(message)s')
    console_handler.setFormatter(formatter)
    LOGGER.addHandler(console_handler)
